/* /////////////////////////////////////////////////////////////////////////////////////////////////
SceneSetup by MacGruber.
https://www.patreon.com/MacGruber_Laboratory

Licensed under CC BY-SA after EarlyAccess ended. (see https://creativecommons.org/licenses/by-sa/4.0/)

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using UnityEngine.XR;
using System.Collections;

namespace MacGruber
{
    public class SceneSetup : MVRScript
    {
		private JSONStorable myImprovedPoV;
		private JSONStorable myPassenger;

		private JSONStorableFloat myYawRotation;
		private JSONStorableFloat myPitchRotation;

		private JSONStorableFloat myBreathingVolumeDesktop;
		private JSONStorableFloat myEffectsVolumeDesktop;
		private JSONStorableFloat myMusicVolumeDesktop;
		private JSONStorableFloat myBreathingVolumeVR;
		private JSONStorableFloat myEffectsVolumeVR;
		private JSONStorableFloat myMusicVolumeVR;

		private UIDynamicButton myStartPossessButton;
		private UIDynamicButton myStartSceneButton;
		private UIDynamicButton myStopPossessButton;

		private bool myNeedsInit = true;
		private bool myActivePossess = false;
		private bool myIsDesktopMode = false;

        public override void Init()
		{
			myNeedsInit = true;
			myIsDesktopMode = !XRSettings.isDeviceActive || string.IsNullOrEmpty(XRSettings.loadedDeviceName);

			SetupInfoText(
				"<color=#606060><size=40><b>Scene Setup</b></size>\nThis scene uses a custom possess system using AcidBubbles <i>ImprovedPoV</i> and <i>Passenger</i> plugins. " +
				"Use the steps on the left to get started.</color>\n\n" +
				"<b>In case you don't want to possess, just skip steps 1 and 2 and press <i>Start Scene</i> directly.</b>\n\n" +
				"<color=#FF0000><b>You can press <i>Space</i> at any time to end possession mode.</b></color>",
				600.0f, true
			);

			SetupInfoOneLine("<size=30><b>Step 1: Possess</b></size>", false);
			myStartPossessButton = SetupButton("Start Possess", StartPossess, false);

			SetupInfoOneLine("<size=30><b>Step 2: Adjust seating position</b></size>", false);
			myYawRotation = SetupSliderFloat("Yaw Rotation (deg)", 0, -180, 180, false);
			myYawRotation.setCallbackFunction += SetRotation;
			myPitchRotation = SetupSliderFloat("Pitch Rotation (deg)", 0, -90, 90, false);
			myPitchRotation.setCallbackFunction += SetRotation;

			SetupInfoOneLine("<size=30><b>Step 3: Adjust audio volume</b></size>", false);
			myBreathingVolumeDesktop = SetupFloat("Breathing Volume Desktop", 0.7f, 0, 1);
			myBreathingVolumeDesktop.setCallbackFunction += SetVolume;
			myEffectsVolumeDesktop = SetupFloat("Effects Volume Desktop", 0.7f, 0, 1);
			myEffectsVolumeDesktop.setCallbackFunction += SetVolume;
			myMusicVolumeDesktop = SetupFloat("Music Volume Desktop", 0.7f, 0, 1);
			myMusicVolumeDesktop.setCallbackFunction += SetVolume;

			myBreathingVolumeVR = SetupFloat("Breathing Volume VR", 0.1f, 0, 1);
			myBreathingVolumeVR.setCallbackFunction += SetVolume;
			myEffectsVolumeVR = SetupFloat("Effects Volume VR", 0.1f, 0, 1);
			myEffectsVolumeVR.setCallbackFunction += SetVolume;
			myMusicVolumeVR = SetupFloat("Music Volume VR", 0.1f, 0, 1);
			myMusicVolumeVR.setCallbackFunction += SetVolume;

			if (myIsDesktopMode)
			{
				CreateSlider(myBreathingVolumeDesktop, false);
				CreateSlider(myEffectsVolumeDesktop, false);
				CreateSlider(myMusicVolumeDesktop, false);
			}
			else
			{
				CreateSlider(myBreathingVolumeVR, false);
				CreateSlider(myEffectsVolumeVR, false);
				CreateSlider(myMusicVolumeVR, false);
			}

			SetupInfoOneLine("<size=30><b>Step 4: Have fun!</b></size>", false);
			myStartSceneButton = SetupButton("Start Scene", StartScene, false);

			SetupInfoOneLine("<size=30><b>Step 5: Stop Possess</b></size>", false);
			myStopPossessButton = SetupButton("Stop Possess", StopPossess, false);
			myStopPossessButton.button.interactable = false;

			SetupAction("OpenSettings", OpenSettings);
			SetupAction("StopPossess", StopPossess);
		}

		private void Update()
		{
			if (myNeedsInit)
			{
				myNeedsInit = false;

				Atom dude = GetAtomById("Dude");
				myImprovedPoV = dude?.GetStorableByID("plugin#0_ImprovedPoV");
				myPassenger = dude?.GetStorableByID("plugin#1_Passenger");

				if (myImprovedPoV == null || myPassenger == null)
				{
					SuperController.LogError("SceneSetup did not find ImprovedPoV and/or Passenger plugin where it expected it.");
					return;
				}

				SetRotation();
				SetVolume();
			}

			if (myActivePossess && Input.GetKey(KeyCode.Space))
			{
				StopPossess();
			}
		}

		private void OpenSettings()
		{
			StartCoroutine(OpenSettingsCO());
		}

		private IEnumerator OpenSettingsCO()
		{
			// Code based on AcidBubbles Keybindings plugin. https://hub.virtamate.com/resources/keybindings.4400/

			// bring up Main HUD
			SuperController sc = SuperController.singleton;
			sc.gameMode = SuperController.GameMode.Edit;
			sc.SelectController(containingAtom.mainController);
			sc.SetActiveUI("SelectedOptions");
			sc.ShowMainHUDAuto();

			// when in VR, move the UI to the controller that pressed the button
			if (sc.isOVR || sc.isOpenVR)
			{
				if (sc.GetLeftSelect())
					sc.MoveMainHUD(sc.isOVR ? sc.touchObjectLeft : sc.viveObjectLeft);
				else if (sc.GetRightSelect())
					sc.MoveMainHUD(sc.isOVR ? sc.touchObjectRight : sc.viveObjectRight);
			}

			// wait for UI to load
			yield return null;

			// bring up Plugin UI
			UITabSelector selector = containingAtom.gameObject.GetComponentInChildren<UITabSelector>(true);
			if (selector == null)
				yield break;
			selector.SetActiveTab("Plugins");

			// bring up _this_ Plugin's UI
			string lastPrefix = null;
			foreach (string uid in containingAtom.GetStorableIDs())
			{
				if (uid == null)
					continue;
				if (!uid.StartsWith("plugin#"))
					continue;
				int prefixEndIndex = uid.IndexOf("_");
				if (prefixEndIndex == -1)
					continue;
				string prefix = uid.Substring(0, prefixEndIndex);
				if (prefix == lastPrefix)
					continue;
				lastPrefix = prefix;
				MVRScript plugin = containingAtom.GetStorableByID(uid) as MVRScript;
				if (plugin != null)
					plugin.UITransform.gameObject.SetActive(plugin == this);
			}
		}

		private void StartPossess()
		{
			if (myImprovedPoV == null || myPassenger == null)
				return;

			myImprovedPoV.SetBoolParamValue("Activate only when possessed", false);
			myPassenger.SetBoolParamValue("Active", true);

			myStartPossessButton.button.interactable = false;
			myStopPossessButton.button.interactable = true;
			EnableAtom("SceneSetup#Text", false);
			EnableAtom("SceneSetup#Open", true);
			EnableAtom("SceneSetup#Stop", true);
			myActivePossess = true;
		}

		private void SetRotation(float _)
		{
			SetRotation();
		}

		private void SetRotation()
		{
			if (myPassenger == null)
				return;

			Quaternion rotation;
			SuperController sc = SuperController.singleton;
			if (sc.isOVR || sc.isOpenVR) // workaround random VaM weirdness
				rotation = Quaternion.Euler(-myPitchRotation.val, 0.0f, 0.0f) * Quaternion.Euler(0.0f, myYawRotation.val, 0.0f);
			else
				rotation = Quaternion.Euler(0.0f, myYawRotation.val, 0.0f) * Quaternion.Euler(-myPitchRotation.val, 0.0f, 0.0f);

			Vector3 euler = rotation.eulerAngles;
			if (euler.x > 180.0f)
				euler.x -= 360.0f;
			if (euler.y > 180.0f)
				euler.y -= 360.0f;
			if (euler.z > 180.0f)
				euler.z -= 360.0f;
			myPassenger.SetFloatParamValue("Rotation X", euler.x);
			myPassenger.SetFloatParamValue("Rotation Y", euler.y);
			myPassenger.SetFloatParamValue("Rotation Z", euler.z);
		}

		private void SetVolume(float _)
		{
			SetVolume();
		}

		private void SetVolume()
		{
			string param = myIsDesktopMode ? "Volume Desktop" : "Volume VR";
			float breathingVolume = myIsDesktopMode ? myBreathingVolumeDesktop.val : myBreathingVolumeVR.val;
			SetParam("GirlA", "plugin#0_MacGruber.AudioAttenuation", param, breathingVolume);
			SetParam("GirlB", "plugin#0_MacGruber.AudioAttenuation", param, breathingVolume);

			float effectsVolume = myIsDesktopMode ? myEffectsVolumeDesktop.val : myEffectsVolumeVR.val;
			SetParam("PenisAudioSource", "plugin#1_MacGruber.AudioAttenuation", param, effectsVolume);

			float musicVolume = myIsDesktopMode ? myMusicVolumeDesktop.val : myMusicVolumeVR.val;
			SetParam("AptSpeaker#Left", "plugin#3_MacGruber.AudioAttenuation", param, musicVolume);
			SetParam("AptSpeaker#Right", "plugin#3_MacGruber.AudioAttenuation", param, musicVolume);
		}

		private void StartScene()
		{
			SuperController sc = SuperController.singleton;
			sc.gameMode = SuperController.GameMode.Play;

			myStartSceneButton.button.interactable = false;
			EnableAtom("SceneSetup#Text", false);
			EnableAtom("SceneSetup#Open", true);
			EnableAtom("SceneSetup#Stop", myActivePossess);

			JSONStorable sceneController = GetAtomById("SceneController")?.GetStorableByID("plugin#0_MacGruber.StateMachine");
			if (sceneController != null)
				sceneController.SetStringParamValue("SwitchState", "Start");
		}

		private void StopPossess()
		{
			if (myImprovedPoV == null || myPassenger == null)
				return;

			myImprovedPoV.SetBoolParamValue("Activate only when possessed", true);
			myPassenger.SetBoolParamValue("Active", false);

			myStartPossessButton.button.interactable = true;
			myStopPossessButton.button.interactable = false;
			EnableAtom("SceneSetup#Open", true);
			EnableAtom("SceneSetup#Stop", false);
			myActivePossess = false;
		}

		private void SetupInfoText(string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamic textfield = CreateTextField(storable, rightSide);
			textfield.height = height;
		}

		private void SetupInfoOneLine(string text, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamicTextField textfield = CreateTextField(storable, rightSide);
			textfield.UItext.alignment = TextAnchor.LowerLeft;
			LayoutElement layout = textfield.GetComponent<LayoutElement>();
			layout.preferredHeight = layout.minHeight = 35;
			textfield.height = 35;
		}

		private JSONStorableFloat SetupFloat(string label, float defaultValue, float minValue, float maxValue)
		{
			JSONStorableFloat storable = new JSONStorableFloat(label, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			RegisterFloat(storable);
			return storable;
		}

		private JSONStorableFloat SetupSliderFloat(string label, float defaultValue, float minValue, float maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(label, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			CreateSlider(storable, rightSide);
			RegisterFloat(storable);
			return storable;
		}

		private UIDynamicButton SetupButton(string label, UnityAction callback, bool rightSide)
		{
			UIDynamicButton button = CreateButton(label, rightSide);
			button.button.onClick.AddListener(callback);
			return button;
		}

		private void SetupAction(string name, JSONStorableAction.ActionCallback callback)
		{
			JSONStorableAction action = new JSONStorableAction(name, callback);
			RegisterAction(action);
		}

		private void EnableAtom(string atom, bool enabled)
		{
			JSONStorable storable = GetAtomById(atom)?.GetStorableByID("AtomControl");
			if (storable != null)
				storable.SetBoolParamValue("on", enabled);
		}

		private void SetParam(string atom, string storable, string param, float val)
		{
			JSONStorable s = GetAtomById(atom)?.GetStorableByID(storable);
			if (s != null)
				s.SetFloatParamValue(param, val);
		}
    }
}
